/**************************************************************
 *
 *
 * $Header:$
 *
 * $Log:$
 *
 * MODULE DESCRIPTION
 *
 *   twogauss - MEx file to do sampling from
 *   a distribution made up of two halves of a
 *   gaussian
 *
 *
 * $Author:$
 *
 *************************************************************/

#include <math.h>
#include <stdlib.h>
#include "mex.h"
#include "twogauss.h"

static  double  pi2 = 2.0 * M_PI;
static  double  tiny = 10 / HUGE;


double twogauss(
double		mun,
double		vn,
double		mup,
double		vp
)
{
  static  double  lpi2;
  double	pneg, ppos, alpha, alpha1;
  double	zn, zp, zna, zpa, stdn, stdp;
  double	intneg, intpos, inttot, sneg, spos;
  int		done = 0;
  double  samp;

  lpi2 = log(pi2);

  pneg = 1.0 / sqrt(pi2 * vn) * exp(-(mun * mun) / (2 * vn));
  ppos = 1.0 / sqrt(pi2 * vp) * exp(-(mup * mup) / (2 * vp));

  /* weight of negative portion */
  alpha = (ppos + tiny) /( ppos+pneg+ 2*tiny);
  alpha1 = 1.0 - alpha;

  /* sample */
  zn = mun / sqrt(vn);
  zp = mup / sqrt(vp);
  zna = sqrt(lpi2- 2*log(alpha+tiny));
  zpa = sqrt(lpi2 - 2*log(alpha1+tiny));
  
  if ( (zna > 2.0 && zp < -2.0) || 
       (zpa > 2.0 && zn > 2.0) || 
       (zn > 2.0  && zp < -2.0)  )
  {
    intneg = alpha * cumnorm(mun, vn);
    intpos= alpha1 * cumnorm(-mup, vp);
    inttot = intneg+intpos;

    if (0.0 == inttot)
    {
      samp=0;
      return samp;
    }

    pneg = intneg/inttot;

    if (drand48()  < pneg)  	/* sample from negative portion  */
    {
      if (zn>1)			/* use tail sampling */
      {
	samp = -samptail(-mun,vn);
      } 
      else        		/* use simple rejection sampling  */
      {
	samp = -samprej(-mun,vn);
      }
    }
    else          		/* sample from positive portion */
    {
      if (zp < -1.0)
      {
	samp = samptail(mup,vp);
      } else		/* use simple rejection sampling  */
      {
	samp=samprej(mup,vp);
      }
    }
  }
  else
  {
    stdn = sqrt(vn);
    stdp = sqrt(vp);

    while (0 == done)
    {
      /* sample from mixture */

      if (drand48() <= alpha)	/* sample from negative portion */
      {
	sneg = mun + randn() * stdn;

	if (sneg<=0)
	{
	  samp = sneg;
	  done=1;
	}
      }
      else
      {
	spos = mup+randn() * stdp;
	    
	if (spos > 0.0)
	{
	  samp = spos;
	  done=1;
	}
      }
    }
  }

  return samp;
}

double samptail(
double	mu,
double	v
)
{
  double samp;
  double tiny, lambda, y, k, accept;
  double pexp0, pgauss0, pexpy, pgaussy;
  int done=0;

  tiny=exp(-600);

  if (mu>0)
    {
      printf("samptail: mu must be negative\n");
      return;
    }

  /* form an exponential distribution bound by looking at the value
     and derivative at zero */

  lambda=-v/(mu+tiny);
  pexp0=1/(lambda+tiny);   /* prob density of exponential at 0 */
  
  /* prob density of Gaussian at 0 */

  k=1/sqrt(pi2*v);
  pgauss0=k*exp(-mu*mu/(2*v));   
					      
  while (!done) {

    /* sample from the exp dist */
    y = -lambda*log(drand48());
    
    /* reject according to relative probabilities */
	
    pgaussy=k*exp(-(y-mu)*(y-mu)/(2*v));
    pexpy=1/lambda*exp(-y/lambda);
    
    if (pexpy*pgauss0==0)
      accept=1;
    else
      accept=pgaussy*pexp0/(pexpy*pgauss0);
    
    if (drand48()<accept)
      done=1;
  }
  samp= y;

  return (samp);
}

double samprej(
double		mu,
double		v
)
{
  double samp;
  double sv;
  int done=0;

  sv=sqrt(v);

  if (mu/sv < -3) {
    printf("warning: samprej should not be used in Gaussian tails\n");
    return;
  }

  while (!done) {
    samp= mu + randn() *sv;
    if (samp>0)
      done=1;
  }
  
  return (samp);

}


double randn()
{
  static int iset=0;
  static double gset;
  double fac,r,v1,v2;
    
  if  (iset == 0) 
  {
    do
    {
      v1 = 2.0 * drand48() - 1.0;
      v2 = 2.0 * drand48() - 1.0;
      r=v1*v1+v2*v2;
    } while (r >= 1.0);
    fac=sqrt(-2.0*log(r)/r);
    gset=v1*fac;
    iset=1;
    return (v2*fac);
  } 
  else
  {
    iset=0;
    return (gset);
  }
}

/*******************************************
 * 
 *  Compute integral from -infinity to mu 
 *  for a normal distribution
 *
 ********************************************/
double cumnorm(
double		mu,
double		v
)
{
  double	v2 = 2.0 * v;
  double	z = mu / sqrt(v2);
  double	lambda, pgauss0;
  double	y;

  if (z < 5.5)
  {
    y= (1 + erf(-z)) / 2.0;
  }
  else
  {
    lambda = v * v /mu;		/* mu is positive */
    pgauss0 = 1.0 /sqrt(pi2 * v)* exp(-mu *mu/ (v2));   

    y = pgauss0 * lambda;
  }

  return (y);
}

/**************************************************************
 *
 *
 * NAME: dotProd
 *
 * DESCRIPTION:
 *
 *  Return dot product of row i and col j of the input matrices
 *
 * RETURNS:
 *
 *  
 *************************************************************/
double dotProd(
Matrix		*x,
int		row,
Matrix		*y,
int		col
)
{
  int		i, l;
  int		xRows, yCols;
  double	*px, *py, ret = 0.0;


  if ( (l = mxGetN(x)) == mxGetM(y))
  {
    xRows = mxGetM(x);
    yCols = mxGetM(y);
    px = mxGetPr(x);
    py = mxGetPr(y);

    for ( i = 0;  i < l ; ++ i)
    {
      ret += px[i*xRows + row] * py[yCols*col + i];
    }
  }
  else
  {
    mexErrMsgTxt("dotProd: Invalid dimensions");
  }

  return ret;
}

/**************************************************************
 *
 *
 * NAME: dotProd
 *
 * DESCRIPTION:
 *
 *  Return dot product of x(:,j)' * y(:,i)
 *
 * RETURNS:
 *
 *  
 *************************************************************/
double transDotProd(
Matrix		*x,
int		row,
Matrix		*y,
int		col
)
{
  int		i, l;
  int		xCols, yCols;
  double	*px, *py, ret = 0.0;


  if ( (l = mxGetM(x)) == mxGetM(y))
  {
    xCols = mxGetM(x);
    yCols = mxGetM(y);
    px = mxGetPr(x);
    py = mxGetPr(y);

    for ( i = 0;  i < l ; ++ i)
    {
      ret += px[row*xCols + i] * py[yCols*col + i];
    }
  }
  else
  {
    mexErrMsgTxt("dotProd: Invalid dimensions");
  }

  return ret;
}

double rect(
double		px,
int		NL
)
{
  int		i;
  double	ret = 0.0;

  if (1 == NL)
  {
    ret = MAX(px, 0.0);
  } 
  else if (2 == NL)
  {
    ret = (double)(px > 0.0);
  }
  else if (3 == NL)
  {
    ret = MAX(px, 0.0);
    ret = MIN(ret, 1.0);
  } 
  else
  {
    mexErrMsgTxt("rect: NL must be 1, 2 or 3");
  }

  return (ret);
}

void rectVect(
double		*xr,
double		*x,
int		n
)
{
  int		i;

  for(i = 0 ; i < n ; ++i)
  {
    xr[i] = rect(x[i], 1);
  }
}

